#!/bin/ksh -p
# $Id: $

function dev_vert_copy
{
    typeset src="${1:?}"
    if [ ! -f "$src" ] ; then
        src=$(echo "$src" | sed 's/dev\/vert-/day\/vert-/')
        if [ ! -f "$src" ] ; then
            echo "no jar at dev or day: $src"
            exit 1
        fi
    fi    
    cp "$src" "${2:?}"
}

function read_properties_file
{
    file_readable "$1"

    typeset X KEY VAL

    cat "$1" | while read X
    do
    
        X=`echo $X | sed -e 's/^[ \t]*//' -e '/^[#!]/d' -e 's/\r//' `
    
        if [ -z "$X" ] ; then
    	    continue
        fi
    
        KEY=`echo $X | sed -n 's/^[ \t]*\([A-Za-z][A-Za-z_0-9]*\)[ \t]*[:=]\{0,1\}[ \t]*\(.*\)$/\1/p' `
    
        if [ -z "$KEY" ]; then
    	    continue
        fi
    
        VAL=`echo $X | sed -n 's/^[ \t]*\([A-Za-z][A-Za-z_0-9]*\)[ \t]*[:=]\{0,1\}[ \t]*\(.*\)$/\2/p' `
    
        eval ${KEY}=\"${VAL}\"
    
    done
}

function file_exists
{
    if [ ! -f "$1" ] ; then
        echo "file not found: \"$1\""
        exit 1
    fi
}

function file_readable
{
    file_exists "$1" "$2"
    
    if [ ! -r "$1" ] ; then
        echo "cannot read: $2\"$1\""
        exit 1
    fi
}   

function file_executable
{
    file_exists "$1" "$2"
    
    if [ ! -x "$1" ] ; then
        echo "cannot execute: $2\"$1\""
        exit 1
    fi
}    

function dir_exists
{
    if [ ! -d "$1" ] ; then
        echo "directory not found: $2\"$1\""
        exit 1
    fi
}

function writeable
{
    if [ ! -w "$1" ] ; then
        echo "cannot write: $2\"$1\""
        exit 1
    fi
}

function dir_writeable
{
    dir_exists "$1" "$2"
    writeable  "$1" "$2"
}

function check_file
{
    file_readable "${1:?}" "$2"
    echo "${1:?}"
}

function check_dir
{
    dir_exists "${1:?}" "$2"
    echo "${1:?}"
}

function init_common_env
{
    PROGRAM_DIR="$(absolute_path ${PROGRAM_NAME:?})"
    PROGRAM_NAME="${PROGRAM_NAME##*(?)/}"
    
    : ${SERVER_HOME:="$(absolute_path ${PROGRAM_DIR:?}/..)"}
    
    dir_writeable "${CONFIG_DIR:="${SERVER_HOME:?}/etc"}"       "CONFIG_DIR="
    dir_writeable "${LOGS_DIR:="${SERVER_HOME:?}/logs"}"        "LOGS_DIR="
    dir_writeable "${ARCHIVE_DIR:="${SERVER_HOME:?}/archive"}"  "ARCHIVE_DIR="
    dir_writeable "${TMPDIR:="${LOGS_DIR:?}"}"                    "TMPDIR="

    if [[ -z "${LIB_DIR}" && -d "${SERVER_HOME:?}/lib" ]] ; then
        LIB_DIR="${SERVER_HOME:?}/lib"
    fi
    
    file_executable "${JAVA:="${JAVA_HOME:?}/bin/java"}"        "JAVA="

    : ${ARCHIVE_COUNT_LIMIT:="30"}
    
    : ${JAVA_HEAP:="4G"}
    : ${JAVA_YOUNG_HEAP:=512M}
    
    : ${DATE_TIME:=$(date +\"%y%b%d_%H%M%S%Z\")}
    : ${ASCII_ENCODING:="ISO8859_1"}
    : ${EBCDIC_ENCODING:="cp1047"}

    : ${YEARCUTOFF:="1926"}
    
    TMP="${TMPDIR:?}/.${PROGRAM_NAME:?}.tmp.$$"
    echo >${TMP:?}
    trap "rm -f ${TMP:?}" 0 1 2 3 15

    if [ -n "${LIB_DIR}" ] ; then
        CLASS_DIRS="${CLASS_DIRS}:${LIB_DIR}"
    fi
    
    CLASSPATH="$CLASSPATH:$(check_dir ${CONFIG_DIR:?} CONFIG_DIR=)"
    
    init_os_env
    init_jvm_env
    
    OPTS="$OPTS $(assertion_opts)"
    OPTS="$OPTS $(encoding_opts)"
}

function init_os_env
{
    RSYNC="cp -fpr"

    if [ $(uname -s) = "AIX" ] ; then

        PSTREE="ps -T"
        ECHO_N="echo"
        IPCS_SED_XPRN='s/^[sm] \{1,\}\([A-Za-z0-9_]\{1,\}\).*/\1/'
    
    else
    
        PSTREE="pstree"
        ECHO_N="echo -e"
        IPCS_SED_XPRN='s/^ *\([A-Za-z0-9_]\{1,\}\).*/\1/'
    
    fi
}

function init_jvm_env
{
    "$JAVA" -version >"$TMP" 2>&1

    if grep -iq "AIX" "$TMP" ; then
    
        : ${PROVIDER_JAR:="${JAVA_HOME:?}/lib/ext/ibmjceprovider.jar"}

		# XXX gencon tests better that optavgpause
        OPTS="$OPTS -Xgcpolicy:gencon"
#        OPTS="$OPTS -Xgcpolicy:optavgpause"

    else 
        
        : ${PROVIDER_JAR:="${JAVA_HOME:?}/lib/ext/sunjce_provider.jar"}
        
        # XXX using concurrent new and old generation GC was good for a large heap
        OPTS="$OPTS -XX:+UseConcMarkSweepGC -XX:+UseParNewGC"
#        OPTS="$OPTS -XX:+UseParallelGC"
        OPTS="$OPTS -XX:+AggressiveOpts"
        
    fi
}

function init_db_env
{
    case "${TDR_JDBC_URL:?}" in
    
    *:db2:*)
    
        CLASSPATH="$CLASSPATH:$(check_file ${DB_HOME:?}/java/db2jcc4.jar)"
        CLASSPATH="$CLASSPATH:$(check_file ${DB_HOME:?}/java/db2jcc_license_cu.jar)"
        ;;
    
    *:oracle:*)
    
        CLASSPATH="$CLASSPATH:$(check_file ${DB_HOME:?}/jdbc/lib/ojdbc7.jar)"
        LIBRARY_PATH="$LIBRARY_PATH:$(check_dir ${DB_HOME:?}/lib)"
        ;;
    
    "")

        ;;
    
    *)
    
        echo "ERROR: TDR_JDBC_URL=$TDR_JDBC_URL"
        exit 1
        ;;
    
    esac
}

function init_es_env
{
    # CLASSPATH is update - this hack is only for D4RAPT44-HF1
    # dir_exists "${LIB_JAR_DIR:="${SERVER_HOME:?}/libjar"}"      "LIB_JAR_DIR="
   if [[ -z "${LIB_JAR_DIR}" && -d "${SERVER_HOME:?}/libjar" ]] ; then
        LIB_JAR_DIR="${SERVER_HOME:?}/libjar"
    fi
 

    if [ -n "${LIB_JAR_DIR}" ] ; then
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/apache-log4j-extras-1.2.17.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/commons-cli-1.3.1.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/compiler-0.8.13.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/compress-lzf-1.0.2.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/elasticsearch-2.3.4.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/guava-18.0.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/HdrHistogram-2.1.6.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/hppc-0.7.1.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/jackson-core-2.6.6.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/jackson-dataformat-cbor-2.6.6.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/jackson-dataformat-smile-2.6.6.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/jackson-dataformat-yaml-2.6.6.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/jna-4.1.0.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/joda-convert-1.2.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/joda-time-2.9.4.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/jsr166e-1.1.0.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/jts-1.13.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/log4j-1.2.17.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/securesm-1.0.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/snakeyaml-1.15.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/spatial4j-0.5.jar)"
	CLASSPATH="$CLASSPATH:$(check_file ${LIB_JAR_DIR:?}/t-digest-3.0.jar)"
    fi

}


function init_mq_env
{
    if [[ -n "$MQ_HOME" ]]; then
      CLASS_DIRS="$CLASS_DIRS:$(check_dir ${MQ_HOME:?}/java/lib)"
      LIBRARY_PATH="$LIBRARY_PATH:$(check_dir ${MQ_HOME:?}/java/lib64)"
    else
      if [[ $JMS_PROVIDER == "MqSeries" ]]; then
        echo "MQ_HOME not set, but you have configured it as your JMS provider.  MQ_HOME needs to be set."
        exit 1
      fi
    fi


}

function init_server_env
{
    init_common_env
    
    : ${TAS_CLASS:="com.sas.finance.fraud.txserver.TAServerMain"}        

    init_running_process_env

    init_mq_env
    init_db_env
    init_es_env

    # Java Cryptography Extension (JCE)
    CLASSPATH="$CLASSPATH:$(check_file ${PROVIDER_JAR:?})"

    if [ -n "${LIB_DIR}" ] ; then
        LIBRARY_PATH="$LIBRARY_PATH:${LIB_DIR:?}"
    fi
    
    # OMR SSL certificate exchange, the /. is intentional
    OPTS="$OPTS -Djava.security.egd=file:/dev/./urandom"
    
    OPTS="$OPTS -Dsas.yearcutoff=${YEARCUTOFF:?}"
    OPTS="$OPTS -Djava.library.path=$LIBRARY_PATH"

    OPTS="$OPTS $(vjr_opts)"
}

function absolute_path
{
    typeset X="$1"
    if [ -f "$X" ] ; then
       X="$(dirname $X)"
    fi
    echo "$(cd $X ; pwd -P)"
}

function assertion_opts
{
    if [ "${MODE:?}" != "PROD" ]; then
        echo "-ea -esa"
    fi
}

function encoding_opts
{
    typeset X="-Dascii.encoding=$ASCII_ENCODING"
    X="$X -Debcdic.encoding=$EBCDIC_ENCODING"
    echo $X
}

function clear_env
{
    MQ_HOME=""
    JAVA_HOME=""
    CLASSPATH=""
    MODE="PROD"
}

function debug_opts
{
	typeset X="-agentlib:jdwp=transport=dt_socket,server=y,suspend=n,address=${DEBUG_PORT:="7779"}"
    X="$X -XX:+HeapDumpOnOutOfMemoryError -XX:HeapDumpPath=${LOGS_DIR:?}"
    echo "$X"
}

function vjr_opts
{
    : ${PICKLIST:="${CONFIG_DIR}/picklist"}

    typeset X="-Djava.system.class.loader=com.sas.app.AppClassLoader"

    X="$X -Dsas.app.launch.config=$(check_file ${PICKLIST:?} PICKLIST=)"

    X="$X -classpath $(check_file ${VJR_HOME:?}/eclipse/plugins/sas.launcher.jar)"
    X="$X -Dsas.app.repository.path=$(check_dir ${VJR_HOME:?}/eclipse)"

    X="$X -Dsas.app.class.dirs=$CLASS_DIRS"
    X="$X -Dsas.app.class.path=$CLASSPATH"

    echo "$X"
}

function init_running_process_env
{
    JPS_TAS=$(ps -e -o pid= -o pgid= -o args= | grep "${JAVA:?}" | grep "${SERVER_HOME:?}" | grep "${TAS_CLASS:?} -start" || :)
    TAS_PID=$( echo "$JPS_TAS" | sed -n 's/^[ \t]*\([0-9]\{1,\}\)[ \t]\{1,\}\([0-9]\{1,\}\)[ \t]\{1,\}\(.*\)/\1/p')
    TAS_PGID=$(echo "$JPS_TAS" | sed -n 's/^[ \t]*\([0-9]\{1,\}\)[ \t]\{1,\}\([0-9]\{1,\}\)[ \t]\{1,\}\(.*\)/\2/p')
    TAS_ARGS=$(echo "$JPS_TAS" | sed -n 's/^[ \t]*\([0-9]\{1,\}\)[ \t]\{1,\}\([0-9]\{1,\}\)[ \t]\{1,\}\(.*\)/\3/p')
 
    JPS_TAS_ESC=$(ps -e -o pid= -o pgid= -o args= | grep "${JAVA:?}" |  grep "${SERVER_HOME:?}" | grep "${TAS_CLASS:?} -esc" || :)
    TAS_ESC_PID=$( echo "$JPS_TAS_ESC" | sed -n 's/^[ \t]*\([0-9]\{1,\}\)[ \t]\{1,\}\([0-9]\{1,\}\)[ \t]\{1,\}\(.*\)/\1/p')
    TAS_ESC_PGID=$(echo "$JPS_TAS_ESC" | sed -n 's/^[ \t]*\([0-9]\{1,\}\)[ \t]\{1,\}\([0-9]\{1,\}\)[ \t]\{1,\}\(.*\)/\2/p')
    TAS_ESC_ARGS=$(echo "$JPS_TAS_ESC" | sed -n 's/^[ \t]*\([0-9]\{1,\}\)[ \t]\{1,\}\([0-9]\{1,\}\)[ \t]\{1,\}\(.*\)/\3/p')

    JPS_TAS_VIYAC=$(ps -e -o pid= -o pgid= -o args= | grep "${JAVA:?}" |  grep "${SERVER_HOME:?}" |  grep "${TAS_CLASS:?} -viyac" || :)
    TAS_VIYAC_PID=$( echo "$JPS_TAS_VIYAC" | sed -n 's/^[ \t]*\([0-9]\{1,\}\)[ \t]\{1,\}\([0-9]\{1,\}\)[ \t]\{1,\}\(.*\)/\1/p')
    TAS_VIYAC_PGID=$(echo "$JPS_TAS_VIYAC" | sed -n 's/^[ \t]*\([0-9]\{1,\}\)[ \t]\{1,\}\([0-9]\{1,\}\)[ \t]\{1,\}\(.*\)/\2/p')
    TAS_VIYAC_ARGS=$(echo "$JPS_TAS_VIYAC" | sed -n 's/^[ \t]*\([0-9]\{1,\}\)[ \t]\{1,\}\([0-9]\{1,\}\)[ \t]\{1,\}\(.*\)/\3/p')

}


function prune_by_count
{ 
    integer CNT=0
    
    for ARCHIVE in $(ls -1t "${ARCHIVE_DIR:?}" | tr '\n' ' ')
    do
        if [ -d "${ARCHIVE_DIR:?}/${ARCHIVE:?}" -a -z "${ARCHIVE#$DT_PATTERN}" ] 
        then
            CNT=$CNT+1
            if [ $CNT -gt ${ARCHIVE_COUNT_LIMIT:?} ] ; then
                rm -Rf "${ARCHIVE_DIR:?}/${ARCHIVE:?}"
            fi
        fi
    done
}

function set_DT
{
    if [ $(uname) = "AIX" ] ; then
    
        LM=$(istat "${1:?}" | grep "Last modified:")
        LM=${LM##*( )}
        
        # Last modified:  Thu Aug 14 09:07:42 EDT 2014

        MONTH=$(echo "$LM" | sed 's/  / /'|sed 's/\t/ /'|cut -f 4 -d " " | tr 'a-z' 'A-Z')
        DAY=$(  echo "$LM" | sed 's/  / /'|sed 's/\t/ /'|cut -f 5 -d " ")
        TIME=$( echo "$LM" | sed 's/  / /'|sed 's/\t/ /'|cut -f 6 -d " " | tr -d ':')
        ZONE=$( echo "$LM" | sed 's/  / /'|sed 's/\t/ /'|cut -f 7 -d " ")
        YEAR=$( echo "$LM" | sed 's/  / /'|sed 's/\t/ /'|cut -f 8 -d " " | cut -c 3-4)
        
        if [ ${#DAY} -eq 1 ] ; then
            DAY="0$DAY"
        fi
        
        if [ -n "$YEAR" ] ; then

            DT="$YEAR$MONTH${DAY}_$TIME$ZONE"
            
            DT_PATTERN="[0-9][0-9][A-Z][A-Z][A-Z][0-9][0-9]_[0-9][0-9][0-9][0-9][0-9][0-9][A-Z][A-Z][A-Z]"
            
        else
    
            # no ZONE, ZONE holds YEAR
            DT="$ZONE$MONTH${DAY}_$TIME"
            
            DT_PATTERN="[0-9][0-9][A-Z][A-Z][A-Z][0-9][0-9]_[0-9][0-9][0-9][0-9][0-9][0-9]"
        
        fi
        
    else
    
        LM=$(stat -c "%y" "${1:?}")
        LM=${LM##*( )}
        
        # 2014-08-14 09:11:55.661492414 -0400
    
        DATE=$(echo $LM | cut -d " " -f 1 | tr -d '-')
        TIME=$(echo $LM | cut -d ' ' -f 2 | cut -d '.' -f 1 | tr -d ':')
        ZONE=$(echo $LM | cut -d ' ' -f 3)
        
        DT="${DATE}_$TIME$ZONE"
        
        DT_PATTERN="[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]_[0-9][0-9][0-9][0-9][0-9][0-9][+-][0-9][0-9][0-9][0-9]"
    
    fi
}


function unknown_command
{
    echo "unknown command: '$1'"
    usage
    exit 1
}
