import socket
import sys
import os
import platform
import struct
import random
import binascii
import time
import logging
import getopt
import signal
import msgapi_pb2
import importlib
import inspect
import struct
import ctypes

#----------------------------------------------------------------- 
#                                                                
# COPYRIGHT (C) 2006-2018 BY SAS INSTITUTE INC.,CARY NC          
#   
# Name:      ose.py                                                             
# Purpose:   OnDemand Scoring Engine
# 
#-----------------------------------------------------------------

    
ode = None
engine = None

def initLogging(optionsDictionary):
    FORMAT = "%(asctime)s.%(msecs)d %(module)s (%(funcName)s) %(levelname)s: %(message)s"
    DATEFMT="%y%j %H:%M:%S" 
    logfile = optionsDictionary['LOGFILE']
    verbose = optionsDictionary['VERBOSE']
    if verbose:
        logging.basicConfig(filename=logfile,level=logging.DEBUG,format=FORMAT,datefmt=DATEFMT)
    else:
        logging.basicConfig(filename=logfile,level=logging.INFO,format=FORMAT,datefmt=DATEFMT)

    logging.info("SAS Python Scoring Engine")
    logging.info("Python " + sys.version)
    logging.info("Command line: " + sys.argv[0] + " " + str(sys.argv[1:]) + "\n")

def getCommandLineArguments():
    optionsDictionary = {}
    optionsDictionary['VERBOSE'] = False
    optionsDictionary['MSGAPI_VERSION'] = '04.04.03'
    try:
        options, remainder = getopt.getopt(
           sys.argv[1:],
           'd:o:l:p:v',
           [
           'directory='    
           'odeclassname=',
           'log=',
           'port=', 
           'verbose',
           'msgapi'
           ])
    except getopt.GetoptError as err:
        print('Error:', err)
        sys.exit(1)

    print('ARGV      :', sys.argv[1:])
    print('OPTIONS   :', options)

    for opt, arg in options:
        print("opt=" + opt + ", arg="+arg)

    for opt, arg in options:
        if opt in ('-l', '--log'):
            logfile = str.strip(arg)
            print("Log file is " + logfile)
            optionsDictionary['LOGFILE'] = logfile
        elif opt in ('-v', '--verbose'):
            optionsDictionary['VERBOSE'] = True
        elif opt in ('-p', '--port'):
            port = arg
            try:
               if port.isdigit():
                   optionsDictionary['PORT'] = int(port)
               else:
                   raise Exception()
            except:
                print("Invalid value for port")
                print(type(port))
                sys.exit(1)                                        
        elif opt in ('-o', '--odeclassname'):
            optionsDictionary['ODECLASSNAME'] = arg
        elif opt in ('-d', '--directory'):
            optionsDictionary['DIRECTORY'] = str.strip(arg)
            if not os.path.isdir(arg):
                raise RuntimeError('Rule Result Directory ('+arg+') does not exist or is not a directory')              
                   
    if 'DIRECTORY' not in optionsDictionary:
        raise RuntimeError('Command line option -d <Rule Result Directory> is required')

    return optionsDictionary

class Deployment:
    def __init__(self, rootMultiOrg, modelUsage):  
        logging.debug('Instantiate deployment '+ rootMultiOrg)
        self.rootMultiOrg = rootMultiOrg
        self.modelUsage = modelUsage

    def init(self, engine):
        self.engine = engine
        
    def score( self, engine, trx):        
        result = None
        for resultIndex in range(1,5):
            for usage in self.modelUsage:
                if usage.resultIndex == resultIndex:
                    if trx.smh_multi_org_node_key.startswith(usage.multi_org_node_key):                                             
                        if usage.model: 
                            result = usage.model.score(trx)
                        if result != None:                            
                            engine.setModelResult(trx, usage.resultIndex, result)  

class Deployments:
    def __init__(self, optionsDictionary):
        self.tenants = set()
        self.models = set()
        self.modelClassNames = set()
        self.engine = None 
        self.optionsDictionary = optionsDictionary

        # Add model package directory to the search path        
        self.modelPackagesDirectory = os.path.join(optionsDictionary['DIRECTORY'],"Models")
        sys.path.append(self.modelPackagesDirectory)          

        optionsDictionary['MODELS'] = self.modelPackagesDirectory

    def addDeployment(self, deployment):        
        self.tenants.add(deployment)
        self.registerModels(deployment)

    def init(self, engine):
        self.engine = engine        
        self.segmentSignatureDictionary = self.buildSegmentSignatureDictionary()

        for deployment in self.tenants:
           deployment.init(engine)
        return None

    def buildSegmentSignatureDictionary(self):

        segmentSignatureDictionary = dict()
      
        # build signature version table 
        # [
        #    'CARD_EG_' : { 'CARD_EG_0100', 'CARD_EG_0200'},
        #    'CARD_DB_' : { 'CARD_DB_0100' }
        # ]
        #
        # First add the signature Content ids associated with the models 
        # that will be scoring transactions. 
        # 
        # Second look to see if there are other versions of the content Id 
        # in the siganture directory and if so, add them to the list.
        #
        contentIds = dict()
        for model in self.models:       
            for signatureDefinition in model.signatures:
                contentId = signatureDefinition[1][:8]
                if contentId not in contentIds:
                    contentIds[contentId] = list()
 
        # Iterate over the signature directory looking for python files 
        # where the python filename's first 8 characters match entries in the contentIds dictionary
        signaturePackageDirectory = os.path.join(self.modelPackagesDirectory,"Signatures")
        signatureFiles = [fn for fn in os.listdir(signaturePackageDirectory) if fn[:8] in contentIds and fn.endswith("py")]
        for signaturefile in signatureFiles:
            contentId = signaturefile[:8]
            contentIdVersion = signaturefile[:12]

            versions = contentIds[contentId]                   
            versions.append(contentIdVersion)

        # print out information 
        logging.debug("Signature Definitions")
        for contentId in contentIds.keys():
            logging.debug("   ContentId " + contentId )
            contentIdVersionList = contentIds[contentId]
            for signatureContentIdVersion in contentIdVersionList:
                logging.debug("      "+ signatureContentIdVersion)

        #   segmentSignatureDictionary structure  
        #   [
        #      'Z00': { CARD_EG_0100, object, [ 'CARD_EG_0100':  object,  'CARD_EG_0200': object], tableId },
        #      'Z01': { CARD_DB_0100, object, [ 'CARD_DB_0100':  object], tableId }
        #   ]
        #   
        for model in self.models:

            for signatureDefinition in model.signatures:

                segmentAcronym = signatureDefinition[0].lower()
                signatureContentIdVersion = signatureDefinition[1]
                signatureContentId = signatureContentIdVersion[:8]

                if not segmentAcronym in segmentSignatureDictionary:
                    values = list()
                    values.append(signatureContentIdVersion)  # [0] output version content id 
                    values.append(None)   # [1] output signature version
                    values.append(dict()) # [2] alternate input versions 
                    values.append(None)   # [3] tableId
                    values.append(list()) # [4] list of MultiOrg Nodes where a more is attached
                                          #     that requires this segment.

                    segmentSignatureDictionary[segmentAcronym] = values

                contentIdVersionList = contentIds[signatureContentId]
                for versionStr in contentIdVersionList:
                    version = int(versionStr[-4:])
                    targetVersion =  int(signatureContentIdVersion[-4:])
        
                    module = self.importModule(self.modelPackagesDirectory,"Signatures", versionStr)

                    klass = getattr(module, versionStr)
                    try:
                        inputSignature = klass(self.optionsDictionary)
                    except:
                        inputSignature = klass()
                        
                    values = segmentSignatureDictionary[segmentAcronym]
                    if targetVersion == version:
                        values[1] = inputSignature
                    values[2][versionStr] = inputSignature

        for deployment in self.tenants:
            for modelUsage in deployment.modelUsage:
                org = modelUsage.multi_org_node_key
                for signatureUsage in modelUsage.model.signatures: 
                    segmentAcronym = signatureUsage[0].lower()
                    values = segmentSignatureDictionary[segmentAcronym]
                    orgs = values[4]
                    orgs.append(org)


        return segmentSignatureDictionary

    def score( self, engine, trx):
        # score transaction
        logging.debug('Score transaction for MO node ' + trx.smh_multi_org_node_key)
        for deployment in self.tenants:
           deployment.score(engine, trx)  
        logging.debug('Done scoring transaction')                   

    def importModule(self, dir, subdir, module):
        directory = os.path.join(dir,subdir)
        self.checkInitFilePresent(directory)
        importName = subdir + "." + module
        module = importlib.import_module(importName)  
        logging.debug("Import " + importName + " = " + str(module))
        return module

    def checkInitFilePresent(self, dir):
       # check and make sure tehre is an __init__.py file in the model package directory. 
        # try and create one if it is not already there.  
        initFile = os.path.join(dir,'__init__.py')     
        if not os.path.isfile(initFile):
           open(initFile, 'a').close()
    
    def registerModel(self, modelClassName):
        if modelClassName in self.modelClassNames:
            return 

        logging.debug('Register Model ' + modelClassName)
        self.modelClassNames.add(modelClassName)       

        instance = None
        modelPackageDirectory = os.path.join(self.modelPackagesDirectory,modelClassName)

        # import all modules in the model pacakge directory
        modelFiles = [fn for fn in os.listdir(modelPackageDirectory) if not fn.startswith('__init__.py') and fn.endswith("py")]
        for modelfile in modelFiles:

            module = self.importModule(self.modelPackagesDirectory,modelClassName, 
                    os.path.splitext(os.path.basename(modelfile))[0])

            # find the classname that matches the package name
            # create an instanance of the model class  
            assert hasattr(module, modelClassName), "Model Class %r was not found in directory %s/%s" % (modelClassName,self.modelPackagesDirectory,modelClassName)
            modelKlass = getattr(module, modelClassName)
            if modelKlass != None:

                try:                   
                    instance = modelKlass(self.optionsDictionary)
                except:
                    logging.debug("try creating model using __init__()")
                    instance = modelKlass()

                if instance != None:
                    logging.debug('Successfully registered model ' + modelClassName)
                    if not hasattr(instance, 'signatures'):
                        logging.debug('Model '+ modelClassName + ' does not use any signatures')
                        instance.signatures = list()
                    else:
                        logging.debug('Model uses the following signatures:')
                    for signatureDefinition in instance.signatures:                      
                        if len(signatureDefinition) == 2:
                           mehTable = signatureDefinition[0][2:]
                        else :
                            mehTable = str(signatureDefinition[2])
                        logging.debug('Signature ' + signatureDefinition[1] + 
                                     ' in transaction segment ' + signatureDefinition[0] +
                                     ' from MEH table ' + mehTable )                        
                        
                    self.models.add(instance)
                else:
                    raise RuntimeError('Unable to register model ' + modelClassName)
        return instance

    def registerModels(self, deployment):
        for modelUsage in deployment.modelUsage:
            modelUsage.model = self.registerModel(modelUsage.modelClassName)       

    def getSignatureUsageInfo(self):
        info = list()
        for deployment in self.tenants:
            for modelUsage in deployment.modelUsage:
                for signatureUsage in modelUsage.model.signatures:
                    segmentAcronym = signatureUsage[0].lower()
                    signatureContentIdVersion = signatureUsage[1]
                    entry = self.segmentSignatureDictionary[segmentAcronym]
                    signature = entry[1]  
                    tableId = entry[3] 
                    if tableId == None:                 
                        tableId = segmentAcronym[1:]
                    keyFieldName = " "                        
                    if hasattr(signature, 'keyFieldName'):
                        keyFieldName =  signature.keyFieldName
                    maxSerializedLength  = 0
                    if hasattr(signature, "maxSerializedLength"):                        
                        maxSerializedLength = signature.maxSerializedLength 
                    else:
                        logging.warning("Signature " + signature.__class__.__name__ + " did not specify a maxSerializedLength")
                    usageText =( modelUsage.multi_org_node_key + ","  +  
                                signatureContentIdVersion + "," + 
                                segmentAcronym.upper() + "," + 
                                str(maxSerializedLength) + "," +
                                keyFieldName + "," +
                                tableId
                    )
                    info.append(usageText)
        return info

class Signature:
    def __init__(self):
        self.persistInd = False
        self.present = False

    def reset(self):
         assert(False), self.__class__.__name__ + " must override reset() method"

    def convert(self, trx, otherSignature):
        assert(False), self.__class__.__name__ + " must override convert() method"

    def deserialize(self, bBlob): 
        assert(False), self.__class__.__name__ + " must override deserialize() method"

    def serialize(self, bBlob): 
        assert(False), self.__class__.__name__ + " must override serialize() method"

    def update(self, trx): 
        assert(False), self.__class__.__name__ + " must override update() method"

class ModelUsage:
    def __init__(self, multi_org_node_key, resultIndex, modelClassName):
        self.multi_org_node_key = multi_org_node_key
        self.resultIndex = resultIndex
        self.modelClassName = modelClassName 
        self.model  = None

class ModelResult:
    """Container for score results
    
       A model result object should be returned by a model's 
       score method if it scored the transaction. If it did not 
       score the transaction, then the model's score method should
       return a null. 
    
       The values in the model result object will be copied into the 
       RRR segment of the returned message.
       
       The RRR segment has 4 sets of returned fields. Which set is 
       populated is chosen when the model is registered in the 
       webapp 

       RRR return fields which are populated from model Result object:
          destination field       source attribute
          rrr_model_id_<n>        model_id
          rrr_model_version_<n>   model_version
          rrr_score_<n>           score
          rrr_model_trace<n>      trace 

    """

    def __init__(self, model, score, *var_args_tuple):
        """Create an model result instance

           Args:
           model: the model creating these results. This class name of
                  model instance will be copied into the RRR_MODEL_ID_<n>
                  and RRR_MODEL_VERSION_<n> fields to identify which model 
                  produced the results. where <n> is the RRR model result set 
                  choosen by the user when they registered the model. 
           score: the score result. This value should be between 000 and 999.
           trace: an optional 4 character error, warning, or diagnostic code 
                  of the model authors choice. 

           Example Usage:
               return ModelResult(modelIsntance, 50)
           or
               return ModelResult(modelInstance, null, '4001')

        """

        self._setScore(model, score)
        if len(var_args_tuple) > 0:
            self.model_trace = var_args_tuple[0]      
    def _setScore(self, model, score):
        self._reset()
        self.score = score
        self.model_id = model.__class__.__name__[0:8]
        self.model_version = model.__class__.__name__[8:13]        
    def _reset(self):
        self.score = None
        self.model_id = "unknown_"
        self.model_version = "????"
        self.model_trace = ""       
        self.sup_score = None 
        self.risk_reason1 = None
        self.risk_reason2 = None
        self.risk_reason3 = None
        self.oper_reason1 = None
        self.oper_reason2 = None
        self.oper_reason3 = None

class Field(object):
    def __init__(self, fieldListEntry):
        self.segmentAcronym = fieldListEntry[0]
        self.name = fieldListEntry[1]
        self.informat = fieldListEntry[2]
        self.startingPosition = fieldListEntry[3]
        self.buildID = None
        self.initValue = None 
        if len(fieldListEntry) == 6 :
           self.buildID = fieldListEntry[4]
           self.initValue = fieldListEntry[5]

class Transaction(object):
    
    localInstanceVariables = {'obj', 'segmentDictionary',  'signatureDictionary', 'segmentSignatureDictionary'}

    def __new__(cls, *args, **kw):
       return super(Transaction,cls).__new__(cls)
  
    def __init__(self,optionsDictionary ):
       
        self.obj = msgapi_pb2.Transaction()
        self.segmentDictionary = dict()        
        self.segmentSignatureDictionary = dict()   

        # Add the message API directory to the search path
        msgApiDirectory = os.path.join(optionsDictionary['DIRECTORY'],"MessageAPI")
        self.load_CIV_Definitions(msgApiDirectory)        

    def __getattribute__(self, attr):
       try:    
            if attr in Transaction.localInstanceVariables:
                return super(Transaction, self).__getattribute__(attr)
            # trx.z00 returns signature object deserialized from Z00 content
            if len(attr) == 3 and attr in self.segmentSignatureDictionary:
                return self.segmentSignatureDictionary[attr][1]

            return object.__getattribute__(self,attr)       
       except AttributeError:
            # trx.smh_source
            return self.obj.__getattribute__(attr[:3]).__getattribute__(attr)

    def __setattr__(self, attr, value): 
       if attr in Transaction.localInstanceVariables:
          super(Transaction, self).__setattr__(attr, value)
       else:
          try:
             if hasattr(self.obj.__getattribute__(attr[:3]),attr):
                self.obj.__getattribute__(attr[:3]).__setattr__(attr, value)
             else:
                super(Transaction, self).__setattr__(attr, value)       
          except AttributeError:
             super(Transaction, self).__setattr__(attr, value)
    
    def __getitem__(self, item):
        return getattr(self, item.lower())

    def load_CIV_Definitions(self, msgApiDirectory):
        
        if not  os.path.isdir(msgApiDirectory):
            logging.warning('Can not find msgApiDirectory directory '+ msgApiDirectory)  
            return 
        sys.path.append(msgApiDirectory) 

        msgApiModule = 'messageAPI'
        msgApiAddendumFile = os.path.join(msgApiDirectory,msgApiModule + ".py")
        if os.path.isfile(msgApiAddendumFile):
            logging.debug('Load Client Input Variable Definitions')
        
            module = importlib.import_module(msgApiModule)   

            # create the user variable attributes from the message API module     
            fieldList = module.__dict__[msgApiModule]
            for fieldListEntry in fieldList:
                field = Field(fieldListEntry)
                if field.segmentAcronym not in self.segmentDictionary:
                    self.segmentDictionary[field.segmentAcronym] = set()
                self.segmentDictionary[field.segmentAcronym ].add(field)
                setattr(self, field.name, None)
        else:
            logging.debug('No Client Input Variables')

    def parseFromString(self, xyz):
        logging.debug('Parse fields from %i bytes' % len(xyz))
        self.obj.ParseFromString(xyz)    
        self.deserialize()
        if self.smh_source=='TRACE':
            logging.info(self.obj.__str__())

        if self.smh_resp_req == '2':
            logging.debug("Resolve Key Request. (smh_resp_req =" +  self.smh_resp_req + ")")
            self.resolveKeys(self)
        else:
            logging.debug("Scoring Request. (smh_resp_req =" +  self.smh_resp_req + ")")            
            self.deserializeModelSignatureSegment()

    # The I##, V##, O## segments are all versioned by build number
    def deserialize(self):
        for segmentAcronym, fields in self.segmentDictionary.items():
            segmentData = self.obj.__getattribute__(segmentAcronym)
            segmentLength = segmentData.__getattribute__(segmentAcronym + '_variable_data_length')
            segmentVarData = segmentData.__getattribute__(segmentAcronym + '_variable_data')                  

            # build id versioned segment 
            if segmentLength > 0:     
                segmentBuildID = segmentData.__getattribute__(segmentAcronym + '_build_num')
                for field in fields:
                    if field.buildID != None and field.buildID > segmentBuildID:
                        if field.informat[0].isdigit():
                            if field.initValue != 'null':
                                setattr(self,field.name,field.initValue)
                            else:
                                setattr(self,field.name,None)
                        else:
                            if field.initValue == '.' or field.initValue == '':
                                setattr(self,field.name,None)
                            else:
                                setattr(self,field.name,int(field.initValue))
                    else:
                        fldlen = 8
                        if field.informat[-1] =='s':
                            fldlen = int(field.informat[:-1])
                        if fldlen+field.startingPosition > segmentLength:
                            logging.error("Segment " + segmentAcronym.upper() + " has invalid length. Field " + field.name + 
                                          " would extend beyond the variable data length(" + str(int(segmentLength)) + ")." )
                            break
                        else:
                            setattr(self,field.name,struct.unpack_from(field.informat, segmentVarData[field.startingPosition:])[0])
                            
    def dump(self, obj):
        logging.debug("Dump of object")
        for attr in dir(obj):
            logging.debug("obj.%s = %r" % (attr, getattr(obj, attr)))

    def resolveKeys(self, trx):
        for segmentAcronym in self.segmentSignatureDictionary.keys():
            values = self.segmentSignatureDictionary[segmentAcronym]
            modelSignature = values[1] # instance of CARD_EG_0100
            
            if not hasattr(modelSignature, 'keyFieldName'):
                multiOrgs = values[4] # org branch this model signature is required for 
                smh_multi_org_node_key = self.smh_multi_org_node_key
                signatureRequired = [org for org in multiOrgs  if smh_multi_org_node_key.startswith(org)]
                if signatureRequired:
                    key = modelSignature.resolve(trx)
                    if key != None:
                        segmentData = self.obj.__getattribute__(segmentAcronym)
                        segmentData.__setattr__(segmentAcronym + "_lookup_key", key)
        return

    def deserializeModelSignatureSegment(self):

        if self.segmentSignatureDictionary == None:
            return

        # For each signature used by a model which is scoring transactions, check to see if the segment associated with 
        # that signature has content and if so parse it and if necessary upgade/downgrade it. If not, clear the signature
        #
        #   TODO: support more than one signature layout for a given segment. 
        #
        #   segmentSignatureDictionary structure  
        #   [
        #      'Z00': { CARD_EG_0100, object, [ 'CARD_EG_0100':  object, None, 'CARD_EG_0200': object, 'Downgrade'] , tableId, ['AA', 'AB']},
        #      'Z01': { CARD_DB_0100, object, [ 'CARD_DB_0100':  object, None] , tableId, ['A']}
        #   ]
        for segmentAcronym in self.segmentSignatureDictionary.keys():
            values = self.segmentSignatureDictionary[segmentAcronym]
            modelContentIdVersion  = values[0]   # CARD_EG_0100
            modelSignature = values[1] # instance of CARD_EG_0100
            modelSignature.persistInd = False

            segmentData = self.obj.__getattribute__(segmentAcronym)
            segmentVarData = segmentData.__getattribute__(segmentAcronym + '_variable_data')            

            inputSegmentContentId = segmentData.__getattribute__(segmentAcronym + '_content_id_version')
            inputSignatureDictionary= values[2]  # {'CARD_EG_0100': object...}
            
            # dont recognize this signature 
            if inputSegmentContentId not in inputSignatureDictionary:
                availableLength = segmentData.__getattribute__(segmentAcronym + '_variable_data_length')
                contentLength = segmentData.__getattribute__(segmentAcronym + '_content_length')
                if segmentVarData == None:
                    logging.debug("Segment "+ segmentAcronym + 
                              " with available content length of " + str(availableLength) +
                              " not present on transaction")
                else:
                    logging.debug("Found segment " + segmentAcronym + 
                              " with available content length of " + str(availableLength) +
                              " , used content length of " + str(contentLength) +
                              " , and unrecognized content version id of '" + inputSegmentContentId +
                              "' of size "+ str(len(segmentVarData)) +". Clear associated  "+ 
                              modelSignature.__class__.__name__ + "signature")               
                inputSignature = modelSignature
                inputSegmentContentId = modelSignature.__class__.__name__
                modelSignature.reset()
            else:
                logging.debug("Found segment " + segmentAcronym +
                        " with expected content '" + inputSegmentContentId + 
                        "' of length " + str(len(segmentVarData)))
                inputSignature = inputSignatureDictionary[inputSegmentContentId]
                inputSignature.deserialize(segmentVarData)

            if  inputSegmentContentId != modelContentIdVersion:
                logging.debug("Convert from "+inputSegmentContentId + " to " +  modelContentIdVersion)
                modelSignature.convert(self, inputSignature)

            if modelSignature.update(self):
                self.serializeModelSignatureSegment(segmentAcronym)                         
            else:
                logging.debug("Do not persist " + segmentAcronym + " content for signature " + inputSignature.__class__.__name__) 

    def serializeToString(self):
        if self.smh_source=='TRACE':
            logging.info(self.obj.__str__())
        return self.obj.SerializeToString()

    def serializeModelSignatureSegment(self, segmentAcronym):    
        signature  = self.segmentSignatureDictionary[segmentAcronym][1]
        segmentData = self.obj.__getattribute__(segmentAcronym)              
        segmentVarData = self.segmentSignatureDictionary[segmentAcronym][3]      
        segmentVarData = signature.serialize()
        usedLength = len(segmentVarData)

        logging.debug("Serialize " + segmentAcronym + " content for signature '" + signature.__class__.__name__ + 
                      "' of length " + str(usedLength)) 

        availableLength = segmentData.__getattribute__(segmentAcronym + '_variable_data_length')
        if  usedLength > availableLength:
            increase = usedLength - availableLength 
            logging.debug("Segment Size increased.. by " + str(increase))

        segmentData.__setattr__(segmentAcronym + '_variable_data_length', usedLength)
        segmentData.__setattr__(segmentAcronym + '_content_length', usedLength)
        segmentData.__setattr__(segmentAcronym + '_content_id_version', signature.__class__.__name__)
        segmentData.__setattr__(segmentAcronym + '_variable_data', segmentVarData)
        segmentData.__setattr__(segmentAcronym + '_persist_ind', 'Y')

class OnDemandScoringEngine:
    def __init__(self, optionsDictionary):  
       logging.debug('OnDemandScoringEngine - Initialize instance') 
       self.ode = None
       self.numTrans = 0
       self.modelResultFields = None
       self.deployments = self.loadDeployments(optionsDictionary)       
       self.deployments.init(self)

    def run(self, ode):
       logging.debug('Starting')
       self.ode = ode
       
       self.sendDeploymentInfo(ode)
       self.initializeModelResultList(ode.transaction)  

       logging.info("Waiting for work...")
       while(True) :
            transaction = self.ode.receive()  
            if transaction == None:
                break

            if transaction.smh_resp_req != '2':
                self.deployments.score(self, transaction)  

            self.ode.reply()                      
            self.numTrans += 1

       logging.debug('Terminating')    

    def sendDeploymentInfo(self, ode):
        logging.debug("Sending Deployment Info...")
        deploymentInfo = self.deployments.getSignatureUsageInfo()
        deploymentInfo.insert(0, "20190204,04.04.03")
        logging.debug(deploymentInfo)
        ode.writeList(deploymentInfo)

    def setModelResult(self, trx, index, result):
        logging.debug('Update transaction with model results['+str(index)+']. (' +
                    "Model = " + result.model_id + result.model_version + 
                    ", score = " + str(result.score) + 
                    ", trace="   + result.model_trace + 
                    ")")
        setattr(trx, 'rrr_model_id_'     +str(index), result.model_id )
        setattr(trx, 'rrr_model_version_'+str(index), result.model_version )
        setattr(trx, 'rrr_score_'        +str(index), result.score )    
        if result.sup_score != None:
            setattr(trx, 'rrr_sup_score_'    +str(index), result.sup_score )
        if result.model_trace != None:
            setattr(trx, 'rrr_model_trace_'  +str(index), result.model_trace )
        if result.risk_reason1 != None:
            setattr(trx, 'rrr_risk_reason1_' +str(index), result.risk_reason1 )
        if result.risk_reason2 != None: 
            setattr(trx, 'rrr_risk_reason2_' +str(index), result.risk_reason2 )
        if result.risk_reason3 != None:
            setattr(trx, 'rrr_risk_reason3_' +str(index), result.risk_reason3 )
        if result.oper_reason1 != None:
            setattr(trx, 'rrr_oper_reason1_' +str(index), result.oper_reason1 )
        if result.oper_reason2 != None:
            setattr(trx, 'rrr_oper_reason2_' +str(index), result.oper_reason2 )
        if result.oper_reason3 != None:
            setattr(trx, 'rrr_oper_reason3_' +str(index), result.oper_reason3 )

    def initializeModelResultList(self, trx):
        self.modelResult_1 = [trx.rrr_model_id_1, trx.rrr_model_version_1, trx.rrr_score_1,
                              trx.rrr_sup_score_1, trx.rrr_model_trace_1, 
                              trx.rrr_risk_reason1_1, trx.rrr_risk_reason2_1, trx.rrr_risk_reason3_1, 
                              trx.rrr_oper_reason1_1, trx.rrr_oper_reason2_1, trx.rrr_oper_reason3_1]
        self.modelResult_2 = [trx.rrr_model_id_2, trx.rrr_model_version_2, trx.rrr_score_2,
                              trx.rrr_sup_score_2, trx.rrr_model_trace_2, 
                              trx.rrr_risk_reason1_2, trx.rrr_risk_reason2_2, trx.rrr_risk_reason3_2, 
                              trx.rrr_oper_reason1_2, trx.rrr_oper_reason2_2, trx.rrr_oper_reason3_2]
        self.modelResult_3 = [trx.rrr_model_id_3, trx.rrr_model_version_3, trx.rrr_score_3,
                              trx.rrr_sup_score_3, trx.rrr_model_trace_3, 
                              trx.rrr_risk_reason1_3, trx.rrr_risk_reason2_3, trx.rrr_risk_reason3_3, 
                              trx.rrr_oper_reason1_3, trx.rrr_oper_reason2_3, trx.rrr_oper_reason3_3]
        self.modelResult_4 = [trx.rrr_model_id_4, trx.rrr_model_version_4, trx.rrr_score_4,
                              trx.rrr_sup_score_4, trx.rrr_model_trace_4, 
                              trx.rrr_risk_reason1_4, trx.rrr_risk_reason2_4, trx.rrr_risk_reason3_4, 
                              trx.rrr_oper_reason1_4, trx.rrr_oper_reason2_4, trx.rrr_oper_reason3_4]
        self.modelResultFields =  [self.modelResult_1, self.modelResult_2, self.modelResult_3, self.modelResult_4 ]    
         
    def loadDeployments( self, optionsDictionary):
        logging.debug('Load Deployment Files')
        deployments = Deployments(optionsDictionary)

        # Add deployment directory to search path
        rulesStudioResultsDirectory = optionsDictionary['DIRECTORY']
        deploymentDirectory = os.path.join(rulesStudioResultsDirectory,"Deployed")
        if not os.path.isdir(deploymentDirectory):
            raise Exception('Can not find deployment directory '+ deploymentDirectory)          
        sys.path.append(deploymentDirectory)

        # add Models directory to search path
        modelDirectory = os.path.join(rulesStudioResultsDirectory,"Models")
        if not os.path.isdir(modelDirectory):
            raise Exception('Can not find model directory '+ modelDirectory)  
        sys.path.append(modelDirectory)  

        # Create instances for each deployment
        deploymentFiles = [fn for fn in os.listdir(deploymentDirectory)  if fn.endswith("py")]
        for deploymentFile in deploymentFiles:
            if  deploymentFile == '__init__.py':
                continue

            # Import Model Package 
            # We requre Model Package to have the same name as model class 
            deploymentModuleName = os.path.splitext(os.path.basename(deploymentFile))[0]
            module = importlib.import_module(deploymentModuleName)    

            # Create an instance of the deployment (ie tenantDeployment = A_Deployment())
            deploymentInstanceName = deploymentModuleName + "_Deployment"
            deployment = getattr(module, deploymentInstanceName)

            deployments.addDeployment(deployment)
        return deployments


class Ode(object):
#    def __new__(cls, *args, **kwds):
#        return object.__new__(cls)
    def __init__(self):
        raise RuntimeError("subclass must override")
    @classmethod         
    def connect(cls, optionsDictionary, transaction):
        raise RuntimeError("subclass must override")
    def receive(self):
        raise RuntimeError("subclass must override")
    def reply(self): 
        raise RuntimeError("subclass must override")
    def close(self):
        raise RuntimeError("subclass must override")

class MockedOde(Ode):
    def __init__(self):
         logging.debug('MockedOde - Initialize instance') 
         self.nTransactions = 0
    @classmethod         
    def connect(cls, optionsDictionary, transaction):
        logging.debug("Connect To ODE")
        self = cls()
        self.transaction = transaction
        self.nTransactions = 0
        return self
    def receive(self):
        logging.debug('Receive transaction from ODE')
        self.transaction.smh_multi_org_node_key = 'AA'
        self.nTransactions = self.nTransactions + 1
        if self.nTransactions > 5:
           raise RuntimeError('Done receiving Transaction')       
        return self.transaction 
    def reply(self): 
        logging.debug('Send response to ODE')
    def close(self):
        logging.debug('Closing socket connection to ODE') 

class SocketOde(Ode):
    def __init__(self):
        logging.debug('SocketOde - Initialize instance')
        #self.recordToFile = open("transaction.bin", "wb")

    @classmethod
    def connect(cls, optionsDictionary, transaction):
        logging.debug("Connect To ODE")
        
        self = cls()

        port = optionsDictionary['PORT']        
        self.transaction = transaction

        # Create a TCP/IP socket
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                
        # Connect the socket to the port where the server is listening
        server_address = ('localhost', port)
        self.sock.connect(server_address)
        logging.info("Connected to %s port %s" % server_address)
        #time.sleep(2)

        return self

    def socket_read_n(self, n):
        """ Read exactly n bytes from the socket.
            Raise RuntimeError if the connection closed before
            n bytes were read.
        """
        buf = b''
        while n > 0:
            data = self.sock.recv(n)
            if data == '':
                raise RuntimeError('unexpected connection close')
            buf +=data
            n -= len(data)
        return buf

    def receive(self):
        try:        

            # Read the transaction length
            bytes = self.socket_read_n(4)
            tranlength = struct.unpack_from('!i', bytes)[0]
            if tranlength < 0 :
                return None
            logging.debug('Receiving %i bytes' % tranlength)
            #self.recordToFile.write(bytes)

            # Read the entire transaction into a string
            transText = self.socket_read_n(tranlength)
            #self.recordToFile.write(transText)                                   

            self.transaction.parseFromString(transText)


        except Exception as e:
            logging.debug("Caught exception :" + str(e),exc_info=True)
            return None

        return self.transaction

    def reply(self): 
        
        transText = self.transaction.serializeToString()    
        logging.debug('Sending %d bytes' % len(transText))    
        numBytes =self.sock.send(struct.pack('!i',len(transText)))
        numBytes = self.sock.send(transText)
        logging.debug('Sent %d bytes' % numBytes)
   
   
    def writeList(self, entries):
        logging.debug('Sending list with %d entries' % len(entries))
        for entry in entries:
            utf8Bytes = entry.encode()
            self.sock.send(struct.pack('!H', len(utf8Bytes)))
            self.sock.send(utf8Bytes)
        self.sock.send(struct.pack('!H', 0))        
        logging.debug('Sent list')
        

    def close(self):
        logging.debug('Closing socket connection to ODE')
        #self.recordToFile.flush()
        #self.recordToFile.close()
        if self.sock != None:
            self.sock.close()
  
def handler_stop_signals(signum, frame):
    print("Inside handler")
    global run
    global ode
    run = False
    if ode != None:
        ode.close()    

def main():
    global ode
    global engine 
    
    optionsDictionary = getCommandLineArguments()
    initLogging(optionsDictionary)   

    try:
        engine = OnDemandScoringEngine(optionsDictionary)    

        transaction = Transaction(optionsDictionary)
        transaction.segmentSignatureDictionary = engine.deployments.segmentSignatureDictionary

        if 'ODECLASSNAME' in optionsDictionary:
            classname = optionsDictionary['ODECLASSNAME']
            klass = globals()[classname]
            ode = klass.connect(optionsDictionary, transaction)
        else:
            ode = SocketOde.connect(optionsDictionary, transaction)
  
        engine.run(ode)          
        
    except Exception as e:
        logging.error("Exception occurred", exc_info=True)
        
    finally:
        logging.info('Python scoring engine shutting down...')
        if ode != None:
           ode.close()
        if engine != None:
           logging.info('Python scoring engine terminated - Trxs = ' + str(engine.numTrans))

        
if __name__ == '__main__':
    main()
 