/*-------------------------------------------------------------------
 * COPYRIGHT (C) 2005 BY SAS INSTITUTE INC.,CARY NC
 *
 * Name: OSE
 * Purpose: Main Entry Point for the OSE engine
 *
 * System Abend Codes:
 * 1 - CMODELS DD Name does not exist or is unusable
 * 2 - Unable to parse Message API metadata
 * 4 - Too many Rule deployment files found
 * 5 - RULES DD name does not exist or is unusable
 * 6 - Message API Version differs between rule deployment files.
 * 9 - Too many models specified to run.
 *10 - no model metadata found in CMODELS for requested model
 *11 - Unsupported version of the Rules deployment file
 *-------------------------------------------------------------------*/
%macro utl_macro_search( macrolib   );
 
   %if (%sysfunc(cexist( &macrolib..sasmacr.RUL_VERSION_CHECK.macro)) EQ 0) %then
   %do;
      %put ERROR: the sasmacr catalog in CRULES location %sysfunc(pathname(&macrolib)) does not exist or is invalid;
      %ABORT RETURN;
   %end;
 
   libname bootstrp (&macrolib);
   option mstored sasmstore=bootstrp;
 
%mend utl_macro_search;
%utl_Macro_Search( CRULES )
 
/*-------------------------------------------------------------------*/
/* Parse command line options :                                      */
/*                                                                   */
/* RPT_VERBOSE    - turns up logging levels from 0 to 13 (default 0) *
/* OSE_SCOREMATCH - override deployment file location and use        */
/*                  <!sasroot>/misc/fraudtranext/a.sas deployment    */
/*                 file executing the model specified in this value  *
/* LOG_MAX_CRITICAL - Number of bad data errors logged before going  */
/*                 silent  (default 1)                               */
/* LOG_MAX_NON_CRITICAL - Number of minor formatting errors logged   */
/*                 before going silent  (default 1)                  */
/*                                                                   */
%Utl_ParseCmdLineArgs;
 

/*-------------------------------------------------------------------*/
/* Include Deployed rule code                                        */
%OSE_Create_OSERules;
  
/*-------------------------------------------------------------------*/
/* Execute pre-DATA step setup for the engine                        */
/*                                                                   */
%global rul_Abort_Flag rul_Abort_Msg;
%OSE_ENVIRONMENT_INITIALIZATION();

options mstored sasmstore=COMPILED;    
data &rpt_diagnostic_dataset;
 
/*-------------------------------------------------------------------*/
/* Stop now if there was an error during initializtion               */
/*                                                                   */
if symget("rul_Abort_Flag") not in ("" "0") then
   %UTL_Complain_And_Die(&rul_abort_msg Error during initialization. Stopping.);
 
/*-------------------------------------------------------------------*/
/* Local Variable Declarations                                       */
/*                                                                   */
Length RPT_verbose 8;    /* Working Variable to control log messages */
Length RPT_rc 8;         /* Working Variable                         */
 
Length RPT_trx_count 8;  /* Working Variable                         */
                         /* Total # of transaction processed by this */
                         /* datastep since its start                 */
Length RPT_Model_CPU_Time 8;   /* Model accumulated CPU time         */
Length RPT_Rules_CPU_Time 8;   /* Rules accumulated CPU time         */
Length RPT_Total_CPU_Time 8;   /* Total accumulated CPU time         */
Length RPT_Init_CPU_Time 8;    /* Initialization CPU time            */
Length RPT_Trx_CPU_Time 8;     /* Transaction CPU time               */
Length RPT_JOBID   $ 20;       /* Our jobname or PID                 */

/*-------------------------------------------------------------------*/
/* Configuration                                                     */
/*                                                                   */
RPT_Verbose = &RPT_VERBOSE;
%XMS_Set_Option("VERBOSEVARNAME", "RPT_Verbose");
%XMS_Set_Option("RULEMACROVERSION", "&RUL_MACRO_VERSION");
%XMS_Set_Option("msgAPIVersion", "&RPT_MsgAPIVersion");
%XMS_Set_Option("MODELCPUTIME", "RPT_Model_CPU_Time");
%XMS_Set_Option("RULESCPUTIME", "RPT_Rules_CPU_Time");
%XMS_Set_Option("TOTALCPUTIME", "RPT_Total_CPU_Time");
 
 
%UTL_LogTime_Begin(TransactionDefs);
%OSE_Transaction_definitions;
%UTL_LogTime_Finish(TransactionDefs);
 
%OSE_Check_MultiOrgVersion;
 
/*-------------------------------------------------------------------*/
/*  One Time Initializations                                         */
/*                                                                   */

RPT_trx_count = 0;
RPT_Model_CPU_Time = 0;
RPT_Rules_CPU_Time = 0;
RPT_Total_CPU_Time = 0;
RPT_JOBID = "&sysjobid";
 
%UTL_LogTime_Begin(Model_Init);
%OSE_model_initialization;
%UTL_LogTime_Finish(Model_Init);
 
%UTL_LogTime_Begin(Rule_Init);
%OSE_Rule_initialization;
%UTL_LogTime_Finish(Rule_Init);
 
%UTL_LogTime_Begin(Create_RuleInfo);
%OSE_Create_RuleInfo;
%UTL_LogTime_Finish(Create_RuleInfo);
 
/*-------------------------------------------------------------------*/
/* Await connection from client                                      */
/*                                                                   */
%XMS_Connect_To_Client();
 
/* Run time error, no retry possible */
if (missing(RPT_rc) or (RPT_rc NE 0)) then do;
   stop;
   end;
 
/*-------------------------------------------------------------------*/
/* Main Transaction Processing Loop...                               */
/*                                                                   */
putlog "Waiting for Message..."; * Do not change (tests use this);
 
 
RPT_done = 0;
%OSE_Set_CPU_Time();
RPT_Init_CPU_Time = RPT_Total_CPU_Time;
RPT_Trx_CPU_Time = 0;
do while (RPT_done EQ 0);
   
   /* Recv transaction and unpack data segments                      */
   %XMS_Receive_From_Client();
 
   /* Intialize Fields used during transaction processing            */
   %OSE_Transaction_Initialization;
 
   /* OSE Controller Error/Directive                                 */
   /*    0x18 - Directive from Controller to shutdown   (decimal 24) */
   /*    0x20 - Abend in Controller PC routine          (decimal 32) */
   /*    1004 - Controller Unreachable                               */
   if ((RPT_RC EQ 018x) or (RPT_RC EQ 020x) or RPT_RC EQ 1004) then do;
      RPT_done = 1;
      continue;
      end;
 
   /* Reload lookup list                                             */
   /*    0x42 - Reload lookup list                      (decimal 66) */
   /*                                                                */
   /* Variable RPT_Lkup_List_Name will be set to the name of the     */
   /* lookup list hash to reload by the controller when the reload   */
   /* is initiated via an operator command.                          */
   else if (RPT_rc EQ 042x) then do;
      %UTL_Reload_Hash_Table();
      RPT_rc = scc_cmd_code;
      %XMS_Reload_Next_OSE();
      continue;
      end;
 
   /* Dump hash                                                      */
   /*                                                                */
   /* Variable RPT_Lkup_List_Name will be set to the name of the     */
   /* lookup list hash to be dumped                                  */
   else if (RPT_rc EQ 054x) then do;
      %UTL_Dump_Hash_Table();
      RPT_rc = scc_cmd_code;
      continue;
      end;
  
   /* Run time error, no retry possible                              */
   /*                                                                */
   /* Possible RPT_rc (RRR_sys_err_code) values                      */
   /*    1000 - No Connection to XMS Server                          */
   /*    1015 - Unrecognized Transaction Segment                     */
   /*    1016 - Unrecognized Transaction                             */
   /*    1017 - Transaction Data Truncated or shorted than expected  */
   /*    1022 - Unrecognized Segment Content Version                 */
   /*                                                                */
   /*                                                                */
   else if (RPT_rc NE 0) then do;
      %OSE_Critical_Error( RPT_rc );
      end;
 
   /* Run time error, no retry possible (RRR_sys_err_code)           */
   /*                                                                */
   /*    1027 - Mismatch between USC MOHD and deployed rules MO      */
   /*    1036 - Mismatch between USC Build # and rules Build #       */
   /*                                                                */
   %OSE_Check_MultiOrgVersion;
   if (RPT_rc NE 0) then do;
      %OSE_Critical_Error( RPT_rc );
      end;
 
   /* Received a system command. Either a ping or a change to        */
   /* Verbose setting                                                */
   /*                                                                */
   else if (SMH_tran_type EQ "CMD" ) then do;
 
      /* Ping Command                                                */
      if (scc_cmd_code EQ 0) then do;
         scc_cmd_info = "Ready for work - " ||
                     "Version 04.04.00 API " ||
                     substr("&RPT_MsgAPIVersion",2,2) || "." ||
                     substr("&RPT_MsgAPIVersion",4,2) || "." ||
                     substr("&RPT_MsgAPIVersion",6,2);
         RPT_Trx_CPU_Time = RPT_Total_CPU_Time - RPT_Init_CPU_Time;
         rtct = RPT_Total_CPU_Time/1000000;
         rmct = RPT_Model_CPU_Time/1000000;
         rrct = RPT_Rules_CPU_Time/1000000;
         rxct = RPT_Trx_CPU_Time/1000000;
         substr(scc_cmd_info,50,12) = put(rtct,time12.2);
         substr(scc_cmd_info,66,12) = put(rmct,time12.2);
         substr(scc_cmd_info,82,12) = put(rrct,time12.2);
         substr(scc_cmd_info,98,12) = put(rxct,time12.2);
         substr(scc_cmd_info,110,12) = put(RPT_trx_count,best.);
         end;
 
      /* Set Verbose Level                                           */
      else  if (scc_cmd_code EQ 1) then do;
         if (upcase(scc_cmd_info) =: "VERBOSE=") then do;
            rpt_verbose = input( substr(scc_cmd_info,9,2) , 2.);
            putlog "Verbose Level set to " rpt_verbose;
            end;
         end;
 
      /* fork command */
      else if (scc_cmd_code EQ 9) then do;
         %xms_fork(scc_cmd_info);
 
         /* if new process, then dont return a response.. */
         if (rpt_rc EQ 0) then do;
            putlog "Waiting for Message..."; * Do not change (tests use this);
            goto RPT_GetNextTransaction;
            end;
 
         /* if parent SAS return proccess ID */
         scc_cmd_info = "SAS process " || trim(put(rpt_rc,8.)) || " created";
         RPT_rc = 0;
         end;
 
      /* else get Build Numbers */
      else if (scc_cmd_code EQ 10) then do;
         scc_cmd_info = %xms_Get_Builds();
         RPT_rc = 0;
         end;
 
      /* else get Log Name */
      else if (scc_cmd_code EQ 11) then do;
         scc_cmd_info = %xms_Get_LogName();
         RPT_rc = 0;
         end;
 
      /* else reload lookup list */
      else if (scc_cmd_code EQ 12) then do;
         RPT_Lkup_List_Name = upcase(scan(scc_cmd_info,1," "));
         %UTL_Reload_Hash_Table();
         if (scc_cmd_code = 0) then do;
            substr(scc_cmd_info,49,9) = put(rpt_hashobs,9.);
            substr(scc_cmd_info,58,9) = put(rpt_hashsize,9.);
            end;
         end;
 
      /* else update/remove lookup list */
      else if (scc_cmd_code EQ 13) or (scc_cmd_code EQ 14) then do;
         RPT_Lkup_List_Name = upcase(scan(scc_cmd_info,1," "));
         %UTL_Update_Hash_Table();
         if (scc_cmd_code = 0) then do;
            substr(scc_cmd_info,49,9) = put(rpt_hashobs,9.);
            substr(scc_cmd_info,58,9) = put(rpt_hashsize,9.);
            end;
         end;
         
      /* else dump lookup list hash */
      else if (scc_cmd_code EQ 15) then do;
         RPT_Lkup_List_Name = upcase(scan(scc_cmd_info,1," "));
         %UTL_Dump_Hash_Table();
         end;
 
      /* Unknown command                                             */
      else
         scc_cmd_info = "Unknown Command Code...";
 
      goto RPT_DoneWithTransaction;
      end;
 
   /* Start of Model related activities. Either Resolving keys or    */
   /* Scoring                                                        */
   %OSE_Set_Timestamp(7);
   %OSE_Set_CPU_Time();
 
   /* Received a request to resolve the model lookup keys            */
   /*                                                                */
   /* Possible RRR_sys_err_code values                               */
   /*    1037 - No Resolve macro found for model                     */
   /*                                                                */
   if (BAND(rptRequestedFeatures, 8000x)) then do;
      smh_resp_req = '0';
      %OSE_Model_resolve_keys;
 
      %OSE_Set_CPU_Time(MODEL);
      goto RPT_DoneWithTransaction;
      end;
 
   /* Successfully received a transaction, up the counter            */
   RPT_trx_count = RPT_trx_count + 1;
   
   /* Caller Requests ODE skip model execution                       */
   if (NOT band(rptRequestedFeatures, 0001x)) then                     
      goto RPT_DoneWithModels;
      
   %OSE_Model_Execution;

RPT_DoneWithModels: 
   %OSE_Set_CPU_Time(MODEL);
   %OSE_Set_Timestamp(8);
   
   /* Caller Requests ODE skip rule execution                        */
   if (NOT band(rptRequestedFeatures, 0002x)) then 
      goto RPT_DoneWithRules_All;
      
   %OSE_User_Signature_Validation;
   %OSE_User_Rule_Execution;
   %OSE_Auth_rule_execution;

RPT_DoneWithRules_Auth: 
   %AGE_Queue_rule_execution;
   
RPT_DoneWithRules_Queue:
   %Done_With_Rules;

RPT_DoneWithRules_All:
   %OSE_Set_CPU_Time(RULES); 
   
RPT_DoneWithTransaction:
   /* set CQV segment length */
   %Mark_Finish_Rules;
      
   /* Send back updated transaction                                  */
   %XMS_Send_To_Client();
 
RPT_GetNextTransaction:
end; /* do while (RPT_done EQ 0) */
 
/* Supported on z/OS only
%OSE_Set_CPU_Time();
rtct = RPT_Total_CPU_Time/1000000;
rmct = RPT_Model_CPU_Time/1000000;
rrct = RPT_Rules_CPU_Time/1000000;
rxct = RPT_Trx_CPU_Time/1000000;
put "NOTE: Transactions Processed: " RPT_trx_count;
put "      OSE CPU Time:           " rtct time12.2;
put "      Transaction CPU Time:   " rxct time12.2;
put "      Model CPU time:         " rmct time12.2;
put "      Rules CPU time:         " rrct time12.2;
*/
put "Total # of transactions Processed: " RPT_trx_count;
rpt_rc = %XMS_Disconnect_From_Client();
stop;

/*-------------------------------------------------------------------*/
/*                                                                   */ 
/* Include Model utility macros if needed                            */
/*                                                                   */ 
%OSE_MODEL_MACROS;

/*-------------------------------------------------------------------*/
/*                                                                   */
/* Score Code                                                        */
/*                                                                   */ 
%OSE_MODEL_LINKS;
 
/*-------------------------------------------------------------------*/
/* Miscellany                                                        */
/*                                                                   */
/*   Upon failure to load a lookup list, terminate engine            */
/*                                                                   */
RPT_ListLoadFailure:
RPT_ListLoadFailure_Release:
   putlog "ERROR: Failure loading lookup list.  Aborting...";
   stop; /* abort abend instead? */
 
 
/*-------------------------------------------------------------------*/
/*                                                                   */
/*   Upon failure to create a hash, terminate engine                 */
/*                                                                   */
RPT_REPORT_BAD_HASH_CREATE:
   msg = catx(" ", "&rul_error_create_mark_hash: RC=", rc, msg);
   putlog "ERROR: " msg;
   stop; /* abort abend instead? */
 
 
/*-------------------------------------------------------------------*/
/*                                                                   */
/*   Upon failure to update a hash, just complain and continue       */
/*                                                                   */
RPT_REPORT_BAD_HASH_UPDATE:
   msg = catx(" ", "&rul_error_update_mark_hash: RC=", rc, msg,
                   "RPT_trx_count=", RPT_trx_count
             );
   %OSE_Warning( %scan(%superq(rul_error_update_mark_hash),
                 1,%str( )), msg=msg );
   return;
 
 
/*-------------------------------------------------------------------*/
/*                                                                   */
/*   Upon highlight area full, just complain and continue            */
/*                                                                   */
RPT_REPORT_MARK_OVERFLOW:
   msg = catx(" ", "&rul_error_update_mark_hash: ", msg,
                   "RPT_trx_count=", RPT_trx_count
             );
   %OSE_Warning( %scan(%superq(rul_error_update_mark_hash),
                 1,%str( )), msg=msg );
   return;
 
 
/*-------------------------------------------------------------------*/
/*                                                                   */
/* Prevent unref label messages (code is not executed)               */
/*                                                                   */
goto RPT_ListLoadFailure;
goto RPT_ListLoadFailure_Release;
goto RPT_DoneWithRules_Queue;
goto RPT_DoneWithRules_Auth;
goto RPT_DoneWithModels;
goto RPT_Report_Bad_Hash_Create;
goto RPT_Report_Bad_Hash_Update;
goto RPT_Report_Mark_Overflow;
goto RPT_DoneWithRules_All; 
 
run;
